/* 
 * net-2-driver for the NI5210 card (i82586 Ethernet chip)
 *
 * This is an extension to the Linux operating system, and is covered by the
 * same Gnu Public License that covers that work. (wow! :-) )
 * 
 * Alphacode 0.4 (94/03/27) for Linux 1.0
 * Copyrights (c) by M.Hipp (mhipp@student.uni-tuebingen.de) 
 * (feel free to mail me on errors!)
 *
 * autoprobe for: base_addr: 0x300,0x280,0x360,0x320,0x340
 *                mem_start: 0xd0000,0xd4000,0xd8000 (8K and 16K)
 *
 * sources: 
 *   the mainsource is the wollman-ie-driver for BSD (including ni52.h)
 *   (though I have rewritten near everything, a lot of comments are original) 
 * other sources:
 *   the crnwyr-driver by Russ Nelson
 *   and (of course) the original drivers (skeleton.c) by Donald Becker
 * (because I never read an Intel 82586 manual, the existing drivers helped
 *  me a lot to understand this tricky chip.)
 */

/*
 * 26.March.94: patches for Linux1.0 and iomem-auto-probe
 *
 * 30.Sep.93: Added nop-chain .. driver now runs with only one Xmit-Buff, too
 */
 
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <asm/bitops.h>
#include <asm/io.h>

#include "dev.h"
#include "eth.h"
#include "skbuff.h"
#include "arp.h"

#include "ni52.h"

#define DEBUG   /* debug on */

/* ni5210 specific */
#define BUS_SIZE   BUS8BIT /* ni5210 has an 8Bit-bus */

#define ni_attn586()  {outb(0,dev->base_addr+NI52_ATTN);}
#define ni_reset586() {outb(0,dev->base_addr+NI52_RESET);}
/*   */

#define make24(ptr) ( (char *) ptr - p->realbase )
#define make16(ptr) ((u_short) (u_long) make24(ptr))

/******************* how to calc the buffers *****************************

sizeof(scp)+2+sizeof(iscp)+sizeof(scb)+sizeof(nop_cmd) = 42 = INIT
sizeof(rfd) = 24; sizeof(rbd) = 12; 
sizeof(tbd) = 8; sizeof(xmit_cmd) = 18; 

examples:
---------

->cfg1: NFRAMES=16, NBUFFS=48, RBUF_SIZE=256, XMITBUFFS=2 ,BUF_LEN=1514

NFRAMES * sizeof(rfd) = 384;
NBUFFS * ( sizeof(rbd) + RBUF_SIZE) = 12864
XMITBUFFS * ( sizeof(tbd+xmit_cmd+nop_cmd) + BUF_LEN) = 3092
INIT = 42
--------------------
16364   (2 bytes left!)

************************

->cfg2: NFRAMES=9, NBUFFS=18, RBUF_SIZE=256, XMITBUFFS=2 ,BUF_LEN=1514

NFRAMES * sizeof(rfd) = 216
NBUFFS * ( sizeof(rbd) + RBUF_SIZE) = 4824
XMITBUFFS * ( sizeof(tbd+xmit_cmd+nop_cmd) + BUF_LEN) = 3092
INIT = 42
------------------
8180    (18 bytes left!)

***************************************************************************/

#if 0
/* config for 16Kram card */
 #define NFRAMES 16		/* number of frames to allow for receive */
 #define NBUFFS 48		/* number of buffers to allocate */
 #define RBUF_SIZE 256	/* size of each buffer, POWER OF 2 & EVEN*/
 #define BUF_LEN 1514	/* length of transmit buffer (EVEN) */
 #define XMITBUFFS 2            /* number of Xmit-Buffs */
#elif 0
/* config-1 for 8Kram card */
 #define NFRAMES 9
 #define NBUFFS 18
 #define RBUF_SIZE 256
 #define BUF_LEN 1514
 #define XMITBUFFS 2
#else
/* config-2 for 8Kram card (test) */
 #define NFRAMES 7
 #define NBUFFS 24
 #define RBUF_SIZE 256
 #define BUF_LEN 1514
 #define XMITBUFFS 1
#endif

#define DELAY(x) {int i=jiffies; \
                  if(loops_per_sec == 1) \
                     while(i+(x)>jiffies); \
                  else \
                     __delay((loops_per_sec>>5)*x); \
                 }

extern void autoirq_setup(int waittime);
extern int autoirq_report(int waittime);
extern void *irq2dev_map[16];

#ifndef HAVE_ALLOC_SKB
#define alloc_skb(size, priority) (struct sk_buff *) kmalloc(size,priority)
#define kfree_skbmem(addr, size) kfree_s(addr,size);
#endif

#ifndef HAVE_PORTRESERVE
#define check_region(ioaddr, size) 		0
#define	snarf_region(ioaddr, size);		do ; while (0)
#endif

#define ETHERCARD_TOTAL_SIZE 16

#define SA_ADDR0 0x02
#define SA_ADDR1 0x07
#define SA_ADDR2 0x01

#undef HAVE_MULTICAST

static int     ni52_probe1(struct device *dev,int ioaddr);
static void    ni52_interrupt(int reg_ptr);
static int     ni52_open(struct device *dev);
static int     ni52_close(struct device *dev);
static int     ni52_send_packet(struct sk_buff *,struct device *);
static struct  enet_statistics *ni52_get_stats(struct device *dev);
#ifdef HAVE_MULTICAST
static void set_multicast_list(struct device *dev, int num_addrs, void *addrs);
#endif

/* helper-functions */
static int     init586(struct device *dev);
static int     check586(struct device *dev,char *where,unsigned size);
static int     alloc586(struct device *dev);
static void    startrecv586(struct device *dev);
static void   *alloc_rfa(struct device *dev,void *ptr);
static void    ni52_tdr(struct device *,volatile struct tdr_cmd_struct *cmd);
static void    ni52_rcv_int(struct device *dev);
static void    ni52_xmt_int(struct device *dev);
static void    ni52_rnr_int(struct device *dev);

struct priv
{
  struct enet_statistics stats;
  unsigned long   realbase;
  volatile int    rfhead,rftail;
  volatile int    rbhead,rbtail;
  volatile struct scp_struct       *scp;  /* volatile is important */
  volatile struct iscp_struct      *iscp; 
  volatile struct scb_struct       *scb;
  volatile struct rfd_struct       *rframes[NFRAMES];
  volatile struct rbd_struct       *rbuffs[NBUFFS];
  volatile char  *cbuffs[NBUFFS];
  volatile struct xmit_buf_struct  *xmit_buffs[XMITBUFFS];
  volatile struct xmit_cmd_struct  *xmit_cmds[XMITBUFFS];
  volatile struct nop_cmd_struct   *nop_cmds[XMITBUFFS+1];
  volatile int    nop_point;
  volatile char  *xmit_cbuffs[XMITBUFFS];
  volatile int    xmit_count,xmit_last;
  volatile int    xmit_on;
};


/**********************************************
 * close device 
 */

static int ni52_close(struct device *dev)
{
  free_irq(dev->irq);
  irq2dev_map[dev->irq] = 0;

  ni_reset586(); /* the hard way to stop the receiver */

  ((struct priv *) dev->priv)->xmit_on = 0;
  dev->start = 0;
  dev->tbusy = 0;

  return 0;
}

/**********************************************
 * open device 
 */

static int ni52_open(struct device *dev)
{
  alloc586(dev);
  init586(dev);  
  startrecv586(dev);

  if(request_irq(dev->irq, &ni52_interrupt)) 
  {    
    ni_reset586();
    return -EAGAIN;
  }  
  irq2dev_map[dev->irq] = dev;

  dev->interrupt = 0;
  dev->tbusy = 0;
  dev->start = 1;

  return 0; /* most done by init */
}

/**********************************************
 * Check to see if there's an 82586 out there. 
 */

static int check586(struct device *dev,char *where,unsigned size)
{
  struct priv *p = (struct priv *) dev->priv;

  p->realbase = (unsigned long) where + size - 0x01000000;

  p->scp = (struct scp_struct *)(p->realbase + SCP_ADDR);
  memset((char *)p->scp,0, sizeof(struct scp_struct));
  
  /*
   * First we put the ISCP at the bottom of memory; this tests to make
   * sure that our idea of the size of memory is the same as the controller's.
   * This is NOT where the ISCP will be in normal operation.
   */
  p->iscp = (struct iscp_struct *)where;
  memset((char *)p->iscp,0, sizeof(struct iscp_struct));

  /* need this to save the pointer for alloc586 */
  p->scb = (struct scb_struct *)where;
  memset((char *)p->scb,0, sizeof(struct scb_struct));

  p->scp->bussize  = BUS_SIZE;
  p->scp->iscp_ptr = make24(p->iscp);

  p->iscp->busy = 1;

  ni_reset586();
  ni_attn586();

  DELAY(2);	/* wait a while... */

  return(p->iscp->busy == 0);
}

/***************************************************+
 * set iscp at the right place , called by check586
 * and open586 and others. 
 */

int alloc586(struct device *dev)
{
  struct priv *p =  (struct priv *) dev->priv; 

  /*
   * relocate the ISCP to its real home, and reset the controller
   */

  p->iscp = (struct iscp_struct *)((char *)p->scp - sizeof(struct iscp_struct));
  memset((char *)p->iscp,0, sizeof(struct iscp_struct));

  memset((char *) p->scp,0,sizeof(struct scp_struct));
  p->scp->iscp_ptr = make24(p->iscp);
  p->scp->bussize = BUS_SIZE;

  p->iscp->busy = 1;

  ni_reset586();
  ni_attn586();

  DELAY(2); 

  if(p->iscp->busy)
    return 0;

  p->iscp->scb_offset = make16(p->scb);
  memset((char *)p->scb,0,sizeof(struct scb_struct));

  return 1;
}

/**********************************************
 * probe the ni5210-card
 */

int ni52_probe(struct device *dev)
{
  int *port, ports[] = {0x300, 0x280, 0x360 , 0x320 , 0x340, 0};
  int base_addr = dev->base_addr;

  if (base_addr > 0x1ff)		/* Check a single specified location. */
    return ni52_probe1(dev, base_addr);
  else if (base_addr > 0)		/* Don't probe at all. */
    return ENXIO;

  for (port = &ports[0]; *port; port++) {
    int ioaddr = *port;
    if (check_region(ioaddr, ETHERCARD_TOTAL_SIZE))
      continue;
    if( !(inb(ioaddr+NI52_REVISION) == NI_REVISION) || 
        !(inb(ioaddr+NI52_ATTRIB) == NI_ATTRIB))
      continue;

    dev->base_addr = ioaddr;
    if (ni52_probe1(dev, ioaddr) == 0)
      return 0;
  }

  dev->base_addr = base_addr;
  return ENODEV;
}

static int ni52_probe1(struct device *dev,int ioaddr)
{
  unsigned char station_addr[6];
  long memaddrs[] = { 0xd0000, 0xd4000, 0xd8000 , 0 };
  int i,size;
  char *memaddr;

  for(i=0;i<ETH_ALEN;i++)
    dev->dev_addr[i] = station_addr[i] = inb(dev->base_addr+i);

  if(station_addr[0] != SA_ADDR0 || station_addr[1] != SA_ADDR1
                                 || station_addr[2] != SA_ADDR2)
    return ENODEV;

  printk("%s (ni52): NetCard at %#3x, ",dev->name,dev->base_addr);

  dev->priv = (void *) kmalloc(sizeof(struct priv),GFP_KERNEL); /* warning: we don't free it on errors */
  memset((char *) dev->priv,0,sizeof(struct priv));

  /* 
   * check (or search) IO-Memory, 8K and 16K
   */
 
  if(dev->mem_start != 0) /* no auto-mem-probe */
  {
    for(i=2;i>0;i--)
    {
      size = i*0x2000;
      if(!check586(dev,(char *) dev->mem_start,size))
        continue;
      if(!alloc586(dev))
        continue;
    }
    if(!i) {
      printk("?Can't find memory at 0x%lx!\n",dev->mem_start);
      return ENODEV;
    }
  }
  else  
  {
    for(i=0;;i++)
    {
      if(memaddrs[i] == 0) {
        printk("memprobe? Can't find io-memory!\n");
        return ENODEV;
      }
      dev->mem_start = memaddrs[i];

      if(!check586(dev,(char *) dev->mem_start,0x2000)) /* check for 8K */
        continue;
      if(!alloc586(dev))
        continue;
 
      memaddr = (char *) dev->mem_start;
      size = 0x2000;
  
      dev->mem_start -= 0x2000;
 
      if(!check586(dev,(char *) dev->mem_start,0x4000)) /* check for 16K */
        break;
      if(!alloc586(dev))
        break;

      memaddr = (char *) dev->mem_start;
      size = 0x4000;
    }
    dev->mem_start = (long) memaddr;
    check586(dev,(char *) dev->mem_start,size); /* final init: (sets p->realbase,p->scp) */
    alloc586(dev);
  }

  printk("Memaddr: 0x%lx, Memsize: %d, ",dev->mem_start,size);

  if(dev->irq < 2)
  {
    autoirq_setup(0);
    ni_reset586();
    ni_attn586();
    if(!(dev->irq = autoirq_report(2)))
    {
      printk("autoirq? Failed to detect IRQ line!\n"); 
      return 1;
    }
  }
  else 
    if(dev->irq == 2) dev->irq = 9;

  printk("IRQ %d.\n",dev->irq);

  init586(dev);

  dev->open           = &ni52_open;
  dev->stop           = &ni52_close;
  dev->get_stats      = &ni52_get_stats;
  dev->hard_start_xmit= &ni52_send_packet;
#ifdef HAVE_MULTICAST
  dev->set_multicast_list = &set_multicast_list;
#endif

  for (i = 0; i < DEV_NUMBUFFS; i++)
    dev->buffs[i] = NULL;

  dev->hard_header	= eth_header;
  dev->add_arp		= eth_add_arp;
  dev->queue_xmit		= dev_queue_xmit;
  dev->rebuild_header	= eth_rebuild_header;
  dev->type_trans		= eth_type_trans;

  dev->type           = ARPHRD_ETHER;
  dev->mtu            = 1500; /* eth_mtu */
  dev->hard_header_len= ETH_HLEN;
  dev->addr_len       = ETH_ALEN;
  for (i = 0; i < ETH_ALEN; i++) {
    dev->broadcast[i]=0xff;
  }

  dev->flags          = IFF_BROADCAST;
  dev->family         = AF_INET;
  dev->pa_addr        = 0;
  dev->pa_brdaddr     = 0;
  dev->pa_mask        = 0;
  dev->pa_alen        = sizeof(unsigned long);

  dev->tbusy = 0;
  dev->interrupt = 0;
  dev->start = 0;
  
  return 0;
}

/********************************************** 
 * init the chip (interrupt should be disabled)
 * needs an correct 'allocated' memory
 */

static int init586(struct device *dev)
{
  void *ptr;
  unsigned long s;
  int i;
  struct priv *p = (struct priv *) dev->priv;
  volatile struct cfg_cmd_struct  *cfg_cmd;
  volatile struct ias_cmd_struct *ias_cmd;

  ptr = (void *) ((char *)p->scb + sizeof(struct scb_struct));

  cfg_cmd = (struct cfg_cmd_struct *)ptr; /* configure-command */
 
  cfg_cmd->cfg_count    = 0x0c;
  cfg_cmd->fifo         = 8;
  cfg_cmd->save_bad     = 0x40;
  cfg_cmd->addr_len     = 0x2e;
  cfg_cmd->priority     = 0;
  cfg_cmd->ifs          = 0x60;
  cfg_cmd->slot_low     = 0;
  cfg_cmd->slot_high    = 0xf2;
  cfg_cmd->promisc      = 0; /* 4=manchaster, 1=promiscous */
  cfg_cmd->crs_cdt      = 0;
  cfg_cmd->min_len      = 60; /* or 64 (60+fcd)*/
  cfg_cmd->junk         = 0xff;
  cfg_cmd->cmd_status      = 0;
  cfg_cmd->cmd_cmd         = CMD_CONFIG | CMD_LAST;
  cfg_cmd->cmd_link        = 0xffff;

  p->scb->cmd_list = make16( cfg_cmd);

  p->scb->cmd = CU_START; /* cmd.-unit start */
  ni_attn586();  /* here: cfg_cmd */
 
  s = jiffies; /* warning: only active with interrupts on !! */
  while(!(cfg_cmd->cmd_status & STAT_COMPL)) if(jiffies-s > 30) break;

  if((cfg_cmd->cmd_status & (STAT_OK|STAT_COMPL)) != (STAT_COMPL|STAT_OK))
  {
    printk("%s (ni52): configure command failed: %x\n",dev->name,cfg_cmd->cmd_status);
    return 1; 
  }

    /*
     * individual address setup
     */
  ias_cmd = (struct ias_cmd_struct *)ptr;

  ias_cmd->cmd_status = 0;
  ias_cmd->cmd_cmd    = CMD_IASETUP | CMD_LAST;
  ias_cmd->cmd_link   = 0xffff;

  memcpy((char *)&ias_cmd->address,(char *) dev->dev_addr,ETH_ALEN);

  p->scb->cmd_list = make16( ias_cmd);

  p->scb->cmd = CU_START; /* cmd.-unit start */
  ni_attn586();

  s = jiffies;
  while(!(ias_cmd->cmd_status & STAT_COMPL)) if(jiffies-s > 30) break;

  if((ias_cmd->cmd_status & (STAT_OK|STAT_COMPL)) != (STAT_OK|STAT_COMPL)) 
  {
    printk("%s (ni52): individual address setup command failed: %04x\n",dev->name,ias_cmd->cmd_status);
    return 1; 
  }

   /* 
    * wire check .. e.g. no resistor e.t.c 
    */
  ni52_tdr(dev,(struct tdr_cmd_struct *)ptr);

   /* 
    * ack interrupts 
    */
  p->scb->cmd = p->scb->status & ST_WHENCE;
  ni_attn586();

  ptr = alloc_rfa(dev,(void *)ptr); /* alloc recv-buffs in a chain */ 

   /*
    * alloc nop-cmd-buffs
    */
  for(i=0;i<XMITBUFFS+1;i++)
  {
    p->nop_cmds[i] = (struct nop_cmd_struct *)ptr;
    p->nop_cmds[i]->cmd_cmd    = 0;
    p->nop_cmds[i]->cmd_status = 0;
    p->nop_cmds[i]->cmd_link   = make16((p->nop_cmds[i]));
    ptr += sizeof(struct nop_cmd_struct);
  }

   /*
    * alloc xmit-buffs 
    */
  for(i=0;i<XMITBUFFS;i++)
  {
    p->xmit_cmds[i] = (struct xmit_cmd_struct *)ptr; /* transmit cmd/buff 0 */
    ptr += sizeof(struct xmit_cmd_struct);
    p->xmit_buffs[i] = (struct xmit_buf_struct *)ptr;
    ptr += sizeof(struct xmit_buf_struct);
    p->xmit_cbuffs[i] = (char *)ptr; /* char-buffs */
    ptr += BUF_LEN;
    if((void *)ptr > (void *)p->iscp) 
    {
      printk("%s (ni52): not enough mem for your configuration!\n",dev->name);
      return 1;
    }   
    memset((char *)(p->xmit_cmds[i]) ,0, sizeof(struct xmit_cmd_struct));
    memset((char *)(p->xmit_buffs[i]),0, sizeof(struct xmit_buf_struct));
    p->xmit_cmds[i]->cmd_status = STAT_COMPL;
    p->xmit_cmds[i]->buffers = make16((p->xmit_buffs[i]));
    p->xmit_buffs[i]->next = 0xffff;
    p->xmit_buffs[i]->buffer = make24((p->xmit_cbuffs[i]));
  }
  
  p->xmit_count = 0; 
  p->xmit_last  = 0;
  p->xmit_on    = 1;
  p->nop_point  = 0;

   /*
    * 'start transmitter' (nop-loop)
    */
  p->scb->cmd_list = make16((p->nop_cmds[0]));
  p->scb->cmd = CU_START;
  ni_attn586();
  while(p->scb->cmd);

  return 0;
}

/******************************************************
 * Here is a helper routine for ni52_nr_int() and init586(). 
 * This sets up the RFA.
 */

static void *alloc_rfa(struct device *dev,void *ptr) 
{
  volatile struct rfd_struct *rfd = (struct rfd_struct *)ptr;
  volatile struct rbd_struct *rbd;
  int i;
  struct priv *p = (struct priv *) dev->priv;

  /* first get frame descriptors */ 

  memset((char *) rfd,0,sizeof(struct rfd_struct)*NFRAMES);
  for(i = 0; i < NFRAMES; i++)
  {
    p->rframes[i] = (rfd + i);
    p->rframes[i]->next = make16((rfd + (i+1) % NFRAMES));
  }
  p->rframes[NFRAMES - 1]->last |= FD_LAST; /* set EOL */
  ptr = (char *) (rfd + NFRAMES);

  /* now get recv-descriptors & recv-buffers */

  rbd = (struct rbd_struct *) ptr;
  ptr += sizeof(struct rbd_struct)*NBUFFS;

  memset((char *) rbd,0,sizeof(struct rbd_struct)*NBUFFS); /* clr descriptors */
  for(i=0;i<NBUFFS;i++)
  {
    p->rbuffs[i] = rbd + i;
    p->rbuffs[i]->next = make16((rbd + (i+1) % NBUFFS));
    p->rbuffs[i]->length = RBUF_SIZE;
    p->rbuffs[i]->buffer = make24(ptr);
    p->cbuffs[i] = (char *)ptr;
    ptr += RBUF_SIZE;
  }
  p->rbuffs[NBUFFS - 1]->length |= RBD_LAST; /* set eol */

  /* We use the head and tail pointers on receive to keep track of
   * the order in which RFDs and RBDs are used. */
  p->rfhead = 0;
  p->rftail = NFRAMES - 1;
  p->rbhead = 0;
  p->rbtail = NBUFFS - 1;

  p->scb->rfd_list = make16((p->rframes[0]));
  p->rframes[0]->buffers = make16( p->rbuffs[0]);

  return ptr;
}

/*********************************************************
 * Run the time-domain reflectometer...
 */

static void ni52_tdr(struct device *dev,volatile struct tdr_cmd_struct *cmd)
{
  int result=0;
  unsigned long s;
  struct priv *p = (struct priv *) dev->priv;

  cmd->cmd_status  = 0;
  cmd->cmd_cmd     = CMD_TDR | CMD_LAST;
  cmd->cmd_link    = 0xffff;
  cmd->tdr_status  = 0;

  p->scb->cmd_list = make16( cmd);

  p->scb->cmd = CU_START; /* cmd.-unit start */
  ni_attn586();

  s = jiffies;
  while(!(cmd->cmd_status & STAT_COMPL))
    if(jiffies - s > 30)
      break;

  DELAY(2); /* wait for result */

  if(cmd->cmd_status & STAT_COMPL)
    result = cmd->tdr_status;
  else
    result = 0x2000 | TDR_TIME;

  p->scb->cmd = p->scb->status & ST_WHENCE;
  ni_attn586(); /* ack the interrupts */

  if(result & TDR_SUCCESS) return;

  if(result & TDR_XCVR) {
    printk("%s: transceiver problem\n",dev->name);
  } else if(result & TDR_OPEN) {
    printk("%s: TDR detected an open %d clocks away\n",dev->name,result & TDR_TIME);
  } else if(result & TDR_SHORT) {
    printk("%s: TDR detected a short %d clocks away\n",dev->name,result & TDR_TIME);
  } else {
    printk("%s: TDR returned unknown status %x\n",dev->name,result);
  }
}

/*********************************************************
 * What to do upon receipt of an interrupt.
 */

static void ni52_interrupt(int reg_ptr)
{
  int irq = -(((struct pt_regs *)reg_ptr)->orig_eax+2);
  struct device *dev = (struct device *) irq2dev_map[irq];
  unsigned short stat;
  int pd = 0;
  struct priv *p = (struct priv *) dev->priv;

  dev->interrupt = 1;

  for(stat=p->scb->status;stat & ST_WHENCE;stat=p->scb->status)
  {
    if(pd) printk("-%04x",(int) stat); /* debug */

    if(stat & (ST_RECV | ST_RNR)) 
      ni52_rcv_int(dev);

    if(stat & ST_DONE)
      ni52_xmt_int(dev);

    if(stat & ST_RNR) 
    {
      printk("%s (ni52): rnr: %04x ",dev->name,(int) stat);
      ni52_rnr_int(dev); 
      pd = 1; /* local debug on */
    }

    p->scb->cmd = p->scb->status & (ST_WHENCE & stat);
    ni_attn586(); /* ack inter. */
   
    if(stat & ST_DONE) 
    {
      dev->tbusy = 0;
      mark_bh(INET_BH);
    }

    dev->interrupt = 0;
  }
}

/*******************************************************
 * receive-interrupt
 */

static void ni52_rcv_int(struct device *dev)
{
  int i,status,head,last,cnt,value;
  unsigned short totlen,sksize,pnt,rbtailold;
  struct sk_buff *skb; 
  struct priv *p = (struct priv *) dev->priv;

  for(i=p->rfhead;;i=(i+1)%NFRAMES)
  {
    status = p->rframes[i]->status;

    if(status & FD_COMPLETE)
    {
      p->rframes[i]->status = 0;
      p->rframes[i]->last |= FD_LAST;
      p->rframes[p->rftail]->last &= ~FD_LAST;
      p->rftail = p->rfhead;
      p->rfhead = (p->rfhead + 1) % NFRAMES;

      if(status & FD_OK)
      {
        totlen = 0; cnt = 0; head=p->rbhead;
        do /* calc recv-frame-len */
        {
          totlen += (value = p->rbuffs[head]->status) & RBD_MASK;
          p->rbuffs[head]->status = 0;
          head++; last = head; head %= NBUFFS; cnt++;
        } while (!(value & RBD_LAST));

        sksize = sizeof(struct sk_buff) + totlen;
        skb = (struct sk_buff *) alloc_skb(sksize, GFP_ATOMIC);

        if (skb != NULL) /* copy header */
        {
          skb->mem_len = sksize;
          skb->mem_addr = skb;
          skb->len = totlen;
          skb->dev = dev;

          if( (last - cnt) < 0)
          {
            pnt = (cnt - last)*RBUF_SIZE;  /* shift possible */
            memcpy( (char *)(skb->data),(char *) p->cbuffs[p->rbhead],pnt);
            memcpy( (char *)(skb->data) + pnt,(char *)p->cbuffs[0],totlen - pnt); 
          }
          else
            memcpy( (char *)(skb->data),(char *)p->cbuffs[p->rbhead],totlen);

          p->rbhead = head;
          rbtailold = p->rbtail; p->rbtail += cnt; p->rbtail %= NBUFFS;
          p->rbuffs[p->rbtail]->length |= RBD_LAST; 
          p->rbuffs[rbtailold]->length &= ~RBD_LAST;

#ifdef HAVE_NETIF_RX
          netif_rx(skb);
          p->stats.rx_packets++;
#else
          skb->lock = 0;
          if(dev_rint((unsigned char *)skb, totlen, IN_SKBUFF, dev)) 
          {
            printk("%s: receive buffers full.\n",dev->name);
            p->stats.rx_dropped++;
          }
          else p->stats.rx_packets++;
#endif
        }
        else
        {
          p->stats.rx_dropped++;
          p->rbhead = head;
          rbtailold = p->rbtail; p->rbtail += cnt; p->rbtail %= NBUFFS;
          p->rbuffs[p->rbtail]->length |= RBD_LAST; 
          p->rbuffs[rbtailold]->length &= ~RBD_LAST;
        }
      }
      else /* frame !(ok) */
      {
        printk("ni52: rfd-error-status: %04x\n",status);
        /* should we here also update the RBD ?! */ 
        p->stats.rx_errors++;
      }
    }
    else break; /* break for-loop */ 
  }
}

/**********************************************************
 * I never got this error , (which should occure if someone 
 * wants to blast your machine) so I couldn't debug it for now.
 * but we _try_ to fix the receiver not ready int.
 */

static void ni52_rnr_int(struct device *dev)
{
  struct priv *p = (struct priv *) dev->priv;

  p->stats.rx_errors++;

  p->scb->cmd = RU_DISABLE;
  ni_attn586(); 
  while(p->scb->cmd);    /* wait for accept cmd. */

  alloc_rfa(dev,(char *)p->rframes[0]);
  startrecv586(dev); /* restart */
}

/**********************************************************
 * handle xmit - interrupt
 */

static void ni52_xmt_int(struct device *dev)
{
  int status;
  struct priv *p = (struct priv *) dev->priv;

  if(!p->xmit_on) return;

  if( (status=p->xmit_cmds[p->xmit_last]->cmd_status) & XS_ERRMASK)
  {
    p->stats.tx_errors++;
    if(status & XS_LATECOLL) 
    {
      printk("%s: late collision\n",dev->name);
      p->stats.collisions++;
    } 
    else if(status & XS_NOCARRIER) 
    {
      p->stats.tx_carrier_errors++;
      printk("%s: no carrier\n",dev->name);
    } 
    else if(status & XS_LOSTCTS) 
    {
      printk("%s: lost CTS\n",dev->name);
    } 
    else if(status & XS_UNDERRUN) 
    {
      printk("%s: DMA underrun\n",dev->name);
    } 
    else if(status & XS_EXCMAX) 
    {
      printk("%s: too many collisions\n",dev->name);
      p->stats.collisions += 16;
    } 
  }
  else 
  {
    p->stats.tx_packets++;
    p->stats.collisions += (status & XS_MAXCOLL);
  }

#ifdef DEBUG
  if(!(p->xmit_cmds[p->xmit_last]->cmd_status & (STAT_COMPL | STAT_OK) )) 
    printk("?");
#endif

  if( (++p->xmit_last) == XMITBUFFS) p->xmit_last = 0;

}

/***********************************************************
 * (re)start the receiver
 */ 

static void startrecv586(struct device *dev)
{
  struct priv *p = (struct priv *) dev->priv;

  p->scb->rfd_list = make16((p->rframes[0]));
  p->scb->cmd = RU_START;
  ni_attn586(); /* start cmd. */
  while(p->scb->cmd); /* wait for accept cmd. */

  DELAY(2); /* isn't necess. */

  p->scb->cmd = p->scb->status & ST_WHENCE;
  ni_attn586();
}

/******************************************************
 * send frame 
 */

static int ni52_send_packet(struct sk_buff *skb, struct device *dev)
{
  int len;
  int next_nop;
  struct priv *p = (struct priv *) dev->priv;

  if(dev->tbusy)
  {
    int tickssofar = jiffies - dev->trans_start;

    if (tickssofar < 10)
      return 1;

#ifdef DEBUG
    printk("%s: xmitter timed out, try to restart! stat: %04x\n",dev->name,p->scb->status);
    printk("%s: command-stats: %04x %04x\n",dev->name,p->xmit_cmds[0]->cmd_status,p->xmit_cmds[1]->cmd_status);
#endif

    ni52_close(dev);
    ni52_open(dev);
    dev->tbusy = 0; /* = 0 ? */
    dev->trans_start = jiffies;
  }

  if(skb == NULL)
  {
    dev_tint(dev);
    return 0;
  }

  if (!skb->arp)
  {
    if(dev->rebuild_header(skb->data, dev)) 
    {
      skb->dev = dev;
      arp_queue (skb);
      return 0; /* arp has queued */
    }
  }
  skb->arp=1;

  if (set_bit(0, (void*)&dev->tbusy) != 0)
     printk("%s: Transmitter access conflict.\n", dev->name);
  else
  {
    memcpy((char *) p->xmit_cbuffs[p->xmit_count],(char *) (skb->data),skb->len);
    len = ETH_ZLEN < skb->len ? skb->len : ETH_ZLEN;

    p->xmit_buffs[p->xmit_count]->flags = XMIT_LAST | len;

    if( (next_nop = p->nop_point + 1) == (XMITBUFFS+1) ) next_nop = 0;

    p->xmit_cmds[p->xmit_count]->cmd_cmd  = CMD_XMIT | CMD_INTR;
    p->xmit_cmds[p->xmit_count]->cmd_status  = 0;
    p->xmit_cmds[p->xmit_count]->cmd_link = p->nop_cmds[next_nop]->cmd_link 
                                          = make16((p->nop_cmds[next_nop]));
    p->nop_cmds[next_nop]->cmd_status = 0;

    p->nop_cmds[p->nop_point]->cmd_link =make16((p->xmit_cmds[p->xmit_count])); 
    dev->trans_start = jiffies;
    p->nop_point = next_nop;
  
    cli();  
    if( (++p->xmit_count) == XMITBUFFS) p->xmit_count = 0;
    if(p->xmit_count != p->xmit_last) dev->tbusy = 0;
    sti();
  }

  if(skb->free) 
    kfree_skb(skb,FREE_WRITE);

  return 0;
}

static struct enet_statistics *ni52_get_stats(struct device *dev)
{
  return &((struct priv *) dev->priv)->stats;
}

 
