/*
 * Intel 82586 Ethernet chip
 * Register, bit, and structure definitions.
 *
 * Main Source:
 * Written by GAW with reference to the Clarkson Packet Driver code for this
 * chip written by Russ Nelson and others.
 */

/* Linuxport by M.Hipp */

/*****************************************************************
 * definitions for NI5210 should be similar(?) on other i82586 card
 */

#define NI52_RESET     0    /* any write here resets the 586 */
#define NI52_ATTN      1    /* any write here sends a Chan attn */
#define NI52_TENA      3    /* 2-5 possibly wrong */
#define NI52_TDIS      2    
#define NI52_INTENA    5   
#define NI52_INTDIS    4   
#define NI52_REVISION  6    /* read here to figure out this board */
#define NI52_ATTRIB    7    /* more information about this board */

#define NI_REVISION 0x00    /* ni5210 value */
#define NI_ATTRIB   0x55    /* ni5210 value */

/* ***************************************************************
 * Standarddefinitons for i82586 chip
 */

#define BUS8BIT  1
#define BUS16BIT 0

/*
 * This is the master configuration block.  It tells the hardware where all
 * the rest of the stuff is.
 */
struct scp_struct
{
  unsigned short mbz;			/* must be zero */
  unsigned char  bussize;		/* true if 8-bit only */
  unsigned char  mbz2[5];		/* must be zero */
  char 		*iscp_ptr;		/* 24-bit physaddr of ISCP */
};

/*
 * Note that this is wired in hardware; the SCP is always located here, no
 * matter what.
 */
#define SCP_ADDR 0xfffff4

/*
 * The tells the hardware where all the rest of the stuff is, too.
 * FIXME: some of these should be re-commented after we figure out their
 * REAL function.
 */
struct iscp_struct 
{
  u_char  busy;			/* zeroed after init */
  u_char  mbz;
  u_short scb_offset;		/* 16-bit physaddr of next struct */
  caddr_t baseaddr;		/* 24-bit physaddr for all 16-bit vars */
};

/*
 * This FINALLY tells the hardware what to do and where to put it.
 */
struct scb_struct
{
  u_short status;	/* status word */
  u_short cmd;		/* command word */
  u_short cmd_list;	/* 16-pointer to command block list */
  u_short rfd_list;	/* 16-pointer to receive frame list */
  u_short crc_err;	/* CRC errors */
  u_short align_err;	/* Alignment errors */
  u_short resource_err;	/* Resource errors */
  u_short overrun_err;	/* Overrun errors */
};

/* Command values */
#define RU_COMMAND	0x0070	/* mask for RU command */
#define RU_NOP		0	/* for completeness */
#define RU_START	0x0010	/* start receive unit command */
#define RU_ENABLE	0x0020	/* enable receiver command */
#define RU_DISABLE	0x0030	/* disable receiver command */
#define RU_ABORT	0x0040	/* abort current receive operation */

#define CU_COMMAND	0x0700	/* mask for CU command */
#define CU_NOP		0	/* included for completeness */
#define CU_START	0x0100	/* do-command command */
#define CU_RESUME	0x0200	/* resume a suspended cmd list */
#define CU_STOP		0x0300	/* SUSPEND was already taken */
#define CU_ABORT	0x0400	/* abort current command */

#define ACK_COMMAND	0xf000	/* mask for ACK command */
#define ACK_CX		0x8000	/* ack ST_DONE */
#define ACK_FR		0x4000	/* ack ST_RECV */
#define ACK_CNA		0x2000	/* ack ST_ALLDONE */
#define ACK_RNR		0x1000	/* ack ST_RNR */

/* Status values */
#define ST_WHENCE	0xf000	/* mask for cause of interrupt */
#define ST_DONE		0x8000	/* command with I bit completed */
#define ST_RECV		0x4000	/* frame received */
#define ST_ALLDONE	0x2000	/* all commands completed */
#define ST_RNR		0x1000	/* receive not ready */

#define CU_STATUS	0x700	/* mask for command unit status */
#define CU_ACTIVE	0x200	/* command unit is active */
#define CU_SUSPEND	0x100	/* command unit is suspended */

#define RU_STATUS	0x70	/* mask for receiver unit status */
#define RU_SUSPEND	0x10	/* receiver is suspended */
#define RU_NOSPACE	0x20	/* receiver has no resources */
#define RU_READY	0x40	/* reveiver is ready */

/*
 * This is filled in partially by the chip, partially by us.
 */
struct rfd_struct 
{
  u_short status;		/* status for this frame */
  u_short last;			/* end of frame list flag */
  u_short next;			/* 16-pointer to next RFD */
  u_short buffers;		/* 16-pointer to list of buffer desc's */
  u_char  dest[6];		/* destination ether */
  u_char  src[6];	 	/* source ether */
  u_short length;		/* 802 length/Ether type */
  u_short mbz;			/* must be zero */
};

#define FD_LAST		0x8000	/* last rfd in list */
#define FD_SUSP		0x4000	/* suspend RU after receipt */

#define FD_COMPLETE	0x8000	/* frame is complete */
#define FD_BUSY		0x4000	/* frame is busy */
#define FD_OK		0x2000	/* frame is bad */
#define FD_RNR		0x0200	/* receiver out of resources here */

/*
 * linked list of buffers...
 */
struct rbd_struct 
{
  u_short status;		/* status for this buffer */
  u_short next;			/* 16-pointer to next RBD */
  caddr_t buffer;		/* 24-pointer to buffer for this RBD */
  u_short length;		/* length of the buffer */
  u_short mbz;			/* must be zero */
};

#define RBD_LAST	0x8000	/* last buffer */
#define RBD_USED	0x4000	/* this buffer has data */
#define RBD_MASK	0x3fff	/* size-mask for length */

#define STAT_COMPL	0x8000	/* command is completed */
#define STAT_BUSY	0x4000	/* command is running now */
#define STAT_OK		0x2000	/* command completed successfully */

#define CMD_NOP		0x0000	/* NOP */
#define CMD_IASETUP	0x0001	/* initial address setup */
#define CMD_CONFIG	0x0002	/* configure command */
#define CMD_MCAST	0x0003	/* multicast setup command */
#define CMD_XMIT	0x0004	/* transmit command */
#define CMD_TDR		0x0005	/* time-domain reflectometer command */
#define CMD_DUMP	0x0006	/* dump command */
#define CMD_DIAGNOSE	0x0007	/* diagnostics command */

#define CMD_LAST	0x8000	/* this is the last command in the list */
#define CMD_SUSPEND	0x4000	/* suspend CU after this command */
#define CMD_INTR	0x2000	/* post an interrupt after completion */

/*
 * This is the command to transmit a frame.
 */
struct xmit_cmd_struct 
{
  u_short cmd_status;		/* status of this command */
  u_short cmd_cmd;		/* command word */
  u_short cmd_link;		/* link to next command */
  u_short buffers;		/* 16-pointer to buffer descriptor */
  u_char  destination[6];       /* destination address */
  u_short length;		/* 802.3 length/Ether type field */
};

#define XS_MAXCOLL  	0x000f	/* number of collisions during transmit */
#define XS_EXCMAX	0x0020	/* exceeded maximum number of collisions */
#define XS_SQE		0x0040	/* SQE positive */
#define XS_DEFERRED	0x0080	/* transmission deferred */
#define XS_UNDERRUN	0x0100	/* DMA underrun */
#define XS_LOSTCTS	0x0200	/* Lost CTS */
#define XS_NOCARRIER	0x0400	/* No Carrier */
#define XS_LATECOLL	0x0800	/* Late collision */

#define XS_ERRMASK   0x0fa0  
/*
 * This is a buffer descriptor for a frame to be transmitted.
 */

struct xmit_buf_struct 
{
  u_short flags;		/* see below */
  u_short next;			/* 16-pointer to next desc. */
  caddr_t buffer;		/* 24-pointer to the actual buffer */
};

#define XMIT_LAST 0x8000	/* this TBD is the last one */
/* The rest of the `flags' word is actually the length. */

/*
 * Multicast setup command.
 */

#define MAXMCAST 50		/* must fit in transmit buffer */

struct mcast_cmd_struct 
{
  u_short cmd_status;		/* status of this command */
  u_short cmd_cmd;		/* command word */
  u_short cmd_link;		/* link to next command */
  u_short mcast_bytes;		/* size (in bytes) of multicast addresses */
  u_char  mcast_addrs[MAXMCAST+1][6];	/* space for them */
};

/*
 * Time Domain Reflectometer command.
 */

struct tdr_cmd_struct 
{
  u_short cmd_status;		/* status of this command */
  u_short cmd_cmd;		/* command word */
  u_short cmd_link;		/* link to next command */
  u_short tdr_status;		/* error bits and time */
};

#define TDR_SUCCESS	0x8000	/* TDR succeeded without error */
#define TDR_XCVR	0x4000	/* detected a transceiver problem */
#define TDR_OPEN	0x2000	/* detected an open */
#define TDR_SHORT	0x1000	/* TDR detected a short */
#define TDR_TIME	0x07ff	/* mask for reflection time */

/*
 * nop command
 */

struct nop_cmd_struct
{
  u_short cmd_status;
  u_short cmd_cmd;
  u_short cmd_link;
};

/*
 * Initial Address Setup command
 */
struct ias_cmd_struct 
{
  u_short cmd_status;		/* status of this command */
  u_short cmd_cmd;		/* command word */
  u_short cmd_link;		/* link to next command */
  u_char  address[6];
};

/*
 * Configuration command 
 */
struct cfg_cmd_struct 
{
  u_short cmd_status;	/* status of this command */
  u_short cmd_cmd;	/* command word */
  u_short cmd_link;	/* link to next command */
  u_char  cfg_count;	/* byte count (0x0c) */
  u_char  fifo;		/* fifo (8) */
  u_char  save_bad;	/* save bad frames (0x40) */
  u_char  addr_len;	/* address length (0x2e) (AL-LOC == 1) */
  u_char  priority;	/* priority and backoff (0x0) */
  u_char  ifs;		/* inter-frame spacing (0x60) */
  u_char  slot_low;	/* slot time, LSB (0x0) */
  u_char  slot_high;	/* slot time, MSN, and retries (0xf2) */
  u_char  promisc;	/* 1 if promiscuous, else 0 */
  u_char  crs_cdt;	/* CSMA/CD parameters (0x0) */
  u_char  min_len;	/* min frame length (0x40) */
  u_char  junk;		/* stuff for 82596 (0xff) */
};
